==> cloud-3.0.0-rc2.tar.gz: cloud/modules/cloud_service_providers/k8s/src/Plugin/cloud/store/K8sCloudStorePlugin.php <==
<?php

namespace Drupal\k8s\Plugin\cloud\store;

use Drupal\cloud\Entity\CloudStoreInterface;
use Drupal\cloud\Plugin\cloud\CloudPluginBase;
use Drupal\cloud\Plugin\cloud\config\CloudConfigPluginManagerInterface;
use Drupal\cloud\Plugin\cloud\store\CloudStorePluginInterface;
use Drupal\cloud\Service\EntityLinkRendererInterface;
use Drupal\Component\Serialization\Yaml;
use Drupal\Component\Uuid\UuidInterface;
use Drupal\Core\Config\ConfigFactoryInterface;
use Drupal\Core\DependencyInjection\ClassResolverInterface;
use Drupal\Core\Entity\EntityInterface;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\File\FileSystemInterface;
use Drupal\Core\Plugin\ContainerFactoryPluginInterface;
use Drupal\Core\Session\AccountProxyInterface;
use Drupal\k8s\Service\K8sServiceInterface;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * K8s Cloud Store plugin.
 */
class K8sCloudStorePlugin extends CloudPluginBase implements CloudStorePluginInterface, ContainerFactoryPluginInterface {

  /**
   * Variable to compare with the time difference.
   *
   * @var int
   */
  private $length = 60 * 60;

  /**
   * The K8s Service.
   *
   * @var \Drupal\k8s\Service\K8sServiceInterface
   */
  protected $k8sService;

  /**
   * The Entity Type Manager.
   *
   * @var \Drupal\Core\Entity\EntityTypeManagerInterface
   */
  protected $entityTypeManager;

  /**
   * The UUID service.
   *
   * @var \Drupal\Component\Uuid\UuidInterface
   */
  protected $uuidService;

  /**
   * Current login user.
   *
   * @var \Drupal\Core\Session\AccountProxyInterface
   */
  protected $currentUser;

  /**
   * The config factory.
   *
   * @var \Drupal\Core\Config\ConfigFactoryInterface
   */
  protected $configFactory;

  /**
   * The cloud service provider plugin manager (CloudConfigPluginManager).
   *
   * @var \Drupal\cloud\Plugin\cloud\config\CloudConfigPluginManagerInterface
   */
  protected $cloudConfigPluginManager;

  /**
   * Entity link renderer object.
   *
   * @var \Drupal\cloud\Service\EntityLinkRendererInterface
   */
  protected $entityLinkRenderer;

  /**
   * File system object.
   *
   * @var \Drupal\Core\File\FileSystemInterface
   */
  protected $fileSystem;

  /**
   * The route match.
   *
   * @var \Drupal\Core\Routing\RouteMatchInterface
   */
  protected $routeMatch;

  /**
   * The class resolver.
   *
   * @var \Drupal\Core\DependencyInjection\ClassResolverInterface
   */
  protected $classResolver;

  /**
   * K8sCloudStorePlugin constructor.
   *
   * @param array $configuration
   *   A configuration array containing information about the plugin instance.
   * @param string $plugin_id
   *   The plugin_id for the plugin instance.
   * @param mixed $plugin_definition
   *   The plugin implementation definition.
   * @param \Drupal\k8s\Service\K8sServiceInterface $k8s_service
   *   The Kubernetes cluster.
   * @param \Drupal\Core\Entity\EntityTypeManagerInterface $entity_type_manager
   *   The entity type manager.
   * @param \Drupal\Component\Uuid\UuidInterface $uuid_service
   *   The uuid service.
   * @param \Drupal\Core\Session\AccountProxyInterface $current_user
   *   Current login user.
   * @param \Drupal\Core\Config\ConfigFactoryInterface $config_factory
   *   The factory for configuration objects.
   * @param \Drupal\cloud\Plugin\cloud\config\CloudConfigPluginManagerInterface $cloud_config_plugin_manager
   *   The cloud service provider plugin manager (CloudConfigPluginManager).
   * @param \Drupal\cloud\Service\EntityLinkRendererInterface $entity_link_renderer
   *   The entity link render service.
   * @param \Drupal\Core\File\FileSystemInterface $file_system
   *   The file system service.
   * @param \Drupal\Core\DependencyInjection\ClassResolverInterface $class_resolver
   *   The class resolver.
   */
  public function __construct(array $configuration,
                              $plugin_id,
                              $plugin_definition,
                              K8sServiceInterface $k8s_service,
                              EntityTypeManagerInterface $entity_type_manager,
                              UuidInterface $uuid_service,
                              AccountProxyInterface $current_user,
                              ConfigFactoryInterface $config_factory,
                              CloudConfigPluginManagerInterface $cloud_config_plugin_manager,
                              EntityLinkRendererInterface $entity_link_renderer,
                              FileSystemInterface $file_system,
                              ClassResolverInterface $class_resolver) {
    parent::__construct($configuration, $plugin_id, $plugin_definition);

    $this->k8sService = $k8s_service;
    $this->entityTypeManager = $entity_type_manager;
    $this->uuidService = $uuid_service;
    $this->currentUser = $current_user;
    $this->configFactory = $config_factory;
    $this->cloudConfigPluginManager = $cloud_config_plugin_manager;
    $this->entityLinkRenderer = $entity_link_renderer;
    $this->fileSystem = $file_system;
    $this->classResolver = $class_resolver;
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container, array $configuration, $plugin_id, $plugin_definition) {
    return new static(
      $configuration,
      $plugin_id,
      $plugin_definition,
      $container->get('k8s'),
      $container->get('entity_type.manager'),
      $container->get('uuid'),
      $container->get('current_user'),
      $container->get('config.factory'),
      $container->get('plugin.manager.cloud_config_plugin'),
      $container->get('entity.link_renderer'),
      $container->get('file_system'),
      $container->get('class_resolver')
    );
  }

  /**
   * {@inheritdoc}
   */
  public function getEntityBundleName(): string {
    return $this->pluginDefinition['entity_bundle'];
  }

  /**
   * {@inheritdoc}
   */
  public function buildListHeader(): array {
    return [];
  }

  /**
   * {@inheritdoc}
   */
  public function buildListRow(CloudStoreInterface $entity): array {
    return [];
  }

  /**
   * Load all entities for a given entity type and bundle.
   *
   * @return \Drupal\Core\Entity\EntityInterface[]
   *   Array of Entity Interface.
   */
  public function loadConfigEntities(): EntityInterface {
    return $this->entityTypeManager->getStorage($this->pluginDefinition['entity_type'])
      ->loadByProperties(['type' => [$this->pluginDefinition['entity_bundle']]]);
  }

  /**
   * Load a cloud service provider (CloudConfig) entity.
   *
   * @param string $entity_bundle
   *   Bundle string.
   *
   * @return array
   *   Entity or empty if there is no entity.
   */
  public function loadConfigEntity($entity_bundle): array {
    $entity = $this->entityTypeManager
      ->getStorage($this->pluginDefinition['entity_type'])
      ->loadByProperties(['entity_bundle' => [$entity_bundle]]);

    if (count($entity) === 1) {
      return array_shift($entity);
    }
    return [];
  }

  /**
   * {@inheritdoc}
   */
  public function updateK8sResourceStoreEntity(): void {
    // Set queue.
    $queue = \Drupal::queue('k8s_update_resources_queue');
    $entities = \Drupal::service('plugin.manager.cloud_config_plugin')->loadConfigEntities('k8s');
    if (empty($entities)) {
      return;
    }

    // Get target entities to be updated.
    foreach ($entities ?: [] as $k8s_cluster) {
      $cloud_context = $k8s_cluster->getCloudContext();

      // To check whether k8s metrics server is installed or not.
      if (!k8s_is_metrics_enabled($cloud_context)) {
        $this->messenger->addWarning($this->t('@message Please install @metrics_server_link into @cloud_context', [
          '@message' => $this->t('Node Resource Store requires K8s Metrics Server.'),
          '@metrics_server_link' => $this->k8sService->getMetricsServerLink($cloud_context),
          '@cloud_context' => $cloud_context,
        ]));
        continue;
      }

      $namespace_entities = $this->entityTypeManager
        ->getStorage('k8s_namespace')
        ->loadByProperties([
          'cloud_context' => $cloud_context,
        ]);
      if (empty($namespace_entities)) {
        continue;
      }

      foreach ($namespace_entities ?: [] as $namespace) {
        // Get resource metrics from each namespace.
        $queue->createItem([
          'k8s_method_name' => 'updateK8sNamespaceResourceStoreEntity',
          'cloud_context' => $cloud_context,
          'namespace' => $namespace->getName(),
          // @todo Ideally, we should use cost_type on worker node(s).
          // However, there is no way to get it right now.
          // Thus, temporary value is put here.
          'cost_type' => 'on_demand_yearly',
        ]);
      }
    }
  }

  /**
   * Extract data based on time range.
   *
   * @param string $cloud_store_type
   *   Cloud store type.
   * @param string $cloud_context
   *   The cloud context.
   * @param string $name
   *   The name of the cost store.
   * @param int $from_time
   *   The unix timestamp of the starting time.
   * @param int $to_time
   *   The unix timestamp of the ending time.
   * @param string $cost_type
   *   The cost type.
   *
   * @return array
   *   The result including resources, cost and unix time.
   */
  public function extract($cloud_store_type, $cloud_context, $name, $from_time, $to_time, $cost_type = NULL): array {
    $storage = $this->entityTypeManager->getStorage('cloud_store');
    $query = $storage->getQuery();
    $query->condition('type', $cloud_store_type);
    if (!empty($cloud_context)) {
      $query->condition('field_cloud_context', $cloud_context);
    }
    if (!empty($name)) {
      $query->condition('name', $name);
    }
    if (!empty($cost_type)) {
      $query->condition('field_cost_type', $cost_type);
    }
    $query->condition('type', $cloud_store_type);
    $query->sort('created');
    $entity_ids = $query->execute();
    $entities = $storage->loadMultiple($entity_ids);

    $result = [];
    $total_cost = 0;
    foreach ($entities ?: [] as $entity) {
      $created_time = $entity->get('created')->value;
      if ($to_time < $created_time || $created_time <= $from_time) {
        continue;
      }
      $resources = Yaml::decode(!empty($entity->get('field_resources')->value) ? $entity->get('field_resources')->value : '');
      $result[] = [
        'resources' => $resources,
        'name' => $entity->getName(),
        'created' => $created_time,
      ];
      if ($entity->hasField('field_cost')) {
        $cost = $entity->get('field_cost')->value;
        $result[count($result) - 1]['cost'] = $cost;
        $total_cost += $cost;
      }
    }
    $result['total_cost'] = $total_cost;
    return $result;
  }

  /**
   * {@inheritdoc}
   */
  public function updateCostStoreEntity(): void {
    $entity_type_manager = $this->entityTypeManager;
    $k8s_service = $this->k8sService;
    // Get target entities to be updated.
    $project_entities = $this->entityTypeManager
      ->getStorage('cloud_project')
      ->loadByProperties([]);

    if (empty($project_entities)) {
      return;
    }

    foreach ($project_entities ?: [] as $project_entity) {
      $label = $project_entity->label();
      $k8s_clusters = $project_entity->get('field_k8s_clusters');

      // Check whether namespace exists or not.
      $namespace_entities = $this->entityTypeManager->getStorage('k8s_namespace')
        ->loadByProperties([
          'name' => $label,
        ]);
      if (empty($namespace_entities)) {
        continue;
      }

      // Get latest entity.
      $store_entities = $this->entityTypeManager->getStorage('cloud_store')
        ->loadByProperties([
          'name' => $label,
          'type' => 'k8s_cost_store',
        ]);
      $store_entity = array_pop($store_entities);

      // To check time difference between current and latest update time
      // in order to decide whether to update or not.
      $latest_time = time() - (60 * 60 * 24);
      if (!empty($store_entity)) {
        $latest_time = $store_entity->get('created')->value;
        $time_diff = time() - $latest_time;

        // If previous update is within the specific range,
        // update process will be skipped.
        if ($time_diff < $this->length) {
          continue;
        }
      }

      $result[$label] = [];
      foreach ($k8s_clusters ?: [] as $k8s_cluster) {
        $cloud_context = $k8s_cluster->value;
        $current_time = floor(time() / $this->length) * $this->length;
        $ids = $this->entityTypeManager
          ->getStorage('cloud_store')
          ->getQuery()
          ->condition('type', 'k8s_namespace_resource_store')
          ->condition('field_cloud_context', $cloud_context)
          ->condition('name', $label)
          ->condition('created', [$latest_time, $current_time], 'BETWEEN')
          ->execute();
        $resource_entities = $this->entityTypeManager->getStorage('cloud_store')
          ->loadMultiple($ids);

        // If no resource entity, it tries to catch the resource info from k8s.
        if (empty($resource_entities)) {
          $namespace_entities = $this->entityTypeManager
            ->getStorage('k8s_namespace')->loadByProperties([
              'cloud_context' => $cloud_context,
              'name' => $label,
            ]);
          foreach ($namespace_entities ?: [] as $namespace_entity) {
            $namespace = $namespace_entity->getName();
            $k8s_service->updateK8sNamespaceResourceStoreEntity($cloud_context, $namespace, 'on_demand_yearly');
          }
          continue;
        }

        // Count cost in specific time range from all k8s clusters.
        foreach ($resource_entities ?: [] as $resource_entity) {
          $created_time = $resource_entity->get('created')->value;
          $time_range = ceil($created_time / $this->length) * $this->length;
          if (!isset($result[$label][$time_range])) {
            $result[$label][$time_range] = [
              'total_cost' => 0,
              'cost_type' => '',
              'count' => 0,
              'resources' => [
                'cpu_usage' => 0,
                'memory_usage' => 0,
                'pod_usage' => 0,
              ],
            ];
          }
          $result[$label][$time_range]['total_cost'] += $resource_entity->get('field_cost')->value;
          $result[$label][$time_range]['cost_type'] = $resource_entity->get('field_cost_type')->value;
          $result[$label][$time_range]['count'] += 1;
          $resources = Yaml::decode($resource_entity->get('field_resources')->value);
          $result[$label][$time_range]['resources']['cpu_usage'] += $resources['cpu_usage'];
          $result[$label][$time_range]['resources']['memory_usage'] += $resources['memory_usage'];
          $result[$label][$time_range]['resources']['pod_usage'] += $resources['pod_usage'];
        }
      }

      // Update cost entity.
      foreach ($result[$label] ?: [] as $key => $value) {
        $avg_cost = 0;
        $cpu_usage_avg = 0;
        $memory_usage_avg = 0;
        $pod_usage_avg = 0;
        if ($value['total_cost'] !== 0) {
          $avg_cost = (float) ($value['total_cost'] / $value['count']);
        }
        if ($value['resources']['cpu_usage'] > 0) {
          $cpu_usage_avg = (float) ($value['resources']['cpu_usage'] / $value['count']);
        }
        if ($value['resources']['memory_usage'] > 0) {
          $memory_usage_avg = (float) ($value['resources']['memory_usage'] / $value['count']);
        }
        if ($value['resources']['pod_usage'] > 0) {
          $pod_usage_avg = (float) ($value['resources']['pod_usage'] / $value['count']);
        }
        $resources = [
          'cpu_usage_avg' => $cpu_usage_avg,
          'memory_usage_avg' => $memory_usage_avg,
          'pod_usage_avg' => $pod_usage_avg,
          'count' => $value['count'],
        ];

        $data = [
          'name' => $label,
          'type' => 'k8s_cost_store',
          'field_cost_type' => $value['cost_type'],
          'field_cost' => $avg_cost,
          'field_resources' => Yaml::encode($resources),
          'created' => $key,
        ];
        // Update cost store entity.
        $entity = $entity_type_manager->getStorage('cloud_store')->create($data);
        $entity->save();
      }
    }
  }

}

==> cloud-3.2.0.tar.gz: cloud/modules/cloud_service_providers/k8s/src/Plugin/cloud/store/K8sCloudStorePlugin.php <==
<?php

namespace Drupal\k8s\Plugin\cloud\store;

use Drupal\cloud\Entity\CloudStoreInterface;
use Drupal\cloud\Plugin\cloud\CloudPluginBase;
use Drupal\cloud\Plugin\cloud\config\CloudConfigPluginManagerInterface;
use Drupal\cloud\Plugin\cloud\store\CloudStorePluginInterface;
use Drupal\cloud\Service\EntityLinkRendererInterface;
use Drupal\Component\Serialization\Yaml;
use Drupal\Component\Uuid\UuidInterface;
use Drupal\Core\Config\ConfigFactoryInterface;
use Drupal\Core\DependencyInjection\ClassResolverInterface;
use Drupal\Core\Entity\EntityInterface;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\File\FileSystemInterface;
use Drupal\Core\Plugin\ContainerFactoryPluginInterface;
use Drupal\Core\Session\AccountProxyInterface;
use Drupal\k8s\Service\K8sServiceInterface;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * K8s Cloud Store plugin.
 */
class K8sCloudStorePlugin extends CloudPluginBase implements CloudStorePluginInterface, ContainerFactoryPluginInterface {

  /**
   * Variable to compare with the time difference.
   *
   * @var int
   */
  private $length = 60 * 60;

  /**
   * The K8s Service.
   *
   * @var \Drupal\k8s\Service\K8sServiceInterface
   */
  protected $k8sService;

  /**
   * The Entity Type Manager.
   *
   * @var \Drupal\Core\Entity\EntityTypeManagerInterface
   */
  protected $entityTypeManager;

  /**
   * The UUID service.
   *
   * @var \Drupal\Component\Uuid\UuidInterface
   */
  protected $uuidService;

  /**
   * Current login user.
   *
   * @var \Drupal\Core\Session\AccountProxyInterface
   */
  protected $currentUser;

  /**
   * The config factory.
   *
   * @var \Drupal\Core\Config\ConfigFactoryInterface
   */
  protected $configFactory;

  /**
   * The cloud service provider plugin manager (CloudConfigPluginManager).
   *
   * @var \Drupal\cloud\Plugin\cloud\config\CloudConfigPluginManagerInterface
   */
  protected $cloudConfigPluginManager;

  /**
   * Entity link renderer object.
   *
   * @var \Drupal\cloud\Service\EntityLinkRendererInterface
   */
  protected $entityLinkRenderer;

  /**
   * File system object.
   *
   * @var \Drupal\Core\File\FileSystemInterface
   */
  protected $fileSystem;

  /**
   * The route match.
   *
   * @var \Drupal\Core\Routing\RouteMatchInterface
   */
  protected $routeMatch;

  /**
   * The class resolver.
   *
   * @var \Drupal\Core\DependencyInjection\ClassResolverInterface
   */
  protected $classResolver;

  /**
   * K8sCloudStorePlugin constructor.
   *
   * @param array $configuration
   *   A configuration array containing information about the plugin instance.
   * @param string $plugin_id
   *   The plugin_id for the plugin instance.
   * @param mixed $plugin_definition
   *   The plugin implementation definition.
   * @param \Drupal\k8s\Service\K8sServiceInterface $k8s_service
   *   The Kubernetes cluster.
   * @param \Drupal\Core\Entity\EntityTypeManagerInterface $entity_type_manager
   *   The entity type manager.
   * @param \Drupal\Component\Uuid\UuidInterface $uuid_service
   *   The uuid service.
   * @param \Drupal\Core\Session\AccountProxyInterface $current_user
   *   Current login user.
   * @param \Drupal\Core\Config\ConfigFactoryInterface $config_factory
   *   The factory for configuration objects.
   * @param \Drupal\cloud\Plugin\cloud\config\CloudConfigPluginManagerInterface $cloud_config_plugin_manager
   *   The cloud service provider plugin manager (CloudConfigPluginManager).
   * @param \Drupal\cloud\Service\EntityLinkRendererInterface $entity_link_renderer
   *   The entity link render service.
   * @param \Drupal\Core\File\FileSystemInterface $file_system
   *   The file system service.
   * @param \Drupal\Core\DependencyInjection\ClassResolverInterface $class_resolver
   *   The class resolver.
   */
  public function __construct(array $configuration,
                              $plugin_id,
                              $plugin_definition,
                              K8sServiceInterface $k8s_service,
                              EntityTypeManagerInterface $entity_type_manager,
                              UuidInterface $uuid_service,
                              AccountProxyInterface $current_user,
                              ConfigFactoryInterface $config_factory,
                              CloudConfigPluginManagerInterface $cloud_config_plugin_manager,
                              EntityLinkRendererInterface $entity_link_renderer,
                              FileSystemInterface $file_system,
                              ClassResolverInterface $class_resolver) {
    parent::__construct($configuration, $plugin_id, $plugin_definition);

    $this->k8sService = $k8s_service;
    $this->entityTypeManager = $entity_type_manager;
    $this->uuidService = $uuid_service;
    $this->currentUser = $current_user;
    $this->configFactory = $config_factory;
    $this->cloudConfigPluginManager = $cloud_config_plugin_manager;
    $this->entityLinkRenderer = $entity_link_renderer;
    $this->fileSystem = $file_system;
    $this->classResolver = $class_resolver;
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container, array $configuration, $plugin_id, $plugin_definition) {
    return new static(
      $configuration,
      $plugin_id,
      $plugin_definition,
      $container->get('k8s'),
      $container->get('entity_type.manager'),
      $container->get('uuid'),
      $container->get('current_user'),
      $container->get('config.factory'),
      $container->get('plugin.manager.cloud_config_plugin'),
      $container->get('entity.link_renderer'),
      $container->get('file_system'),
      $container->get('class_resolver')
    );
  }

  /**
   * {@inheritdoc}
   */
  public function getEntityBundleName(): string {
    return $this->pluginDefinition['entity_bundle'];
  }

  /**
   * {@inheritdoc}
   */
  public function buildListHeader(): array {
    return [];
  }

  /**
   * {@inheritdoc}
   */
  public function buildListRow(CloudStoreInterface $entity): array {
    return [];
  }

  /**
   * Load all entities for a given entity type and bundle.
   *
   * @return \Drupal\Core\Entity\EntityInterface[]
   *   Array of Entity Interface.
   */
  public function loadConfigEntities(): EntityInterface {
    return $this->entityTypeManager->getStorage($this->pluginDefinition['entity_type'])
      ->loadByProperties(['type' => [$this->pluginDefinition['entity_bundle']]]);
  }

  /**
   * Load a cloud service provider (CloudConfig) entity.
   *
   * @param string $entity_bundle
   *   Bundle string.
   *
   * @return array
   *   Entity or empty if there is no entity.
   */
  public function loadConfigEntity($entity_bundle): array {
    $entity = $this->entityTypeManager
      ->getStorage($this->pluginDefinition['entity_type'])
      ->loadByProperties(['entity_bundle' => [$entity_bundle]]);

    if (count($entity) === 1) {
      return array_shift($entity);
    }
    return [];
  }

  /**
   * {@inheritdoc}
   */
  public function updateK8sResourceStoreEntity(): void {
    // Set queue.
    $queue = \Drupal::queue('k8s_update_resources_queue');
    $entities = \Drupal::service('plugin.manager.cloud_config_plugin')->loadConfigEntities('k8s');
    if (empty($entities)) {
      return;
    }

    // Get target entities to be updated.
    foreach ($entities ?: [] as $k8s_cluster) {
      $cloud_context = $k8s_cluster->getCloudContext();

      // To check whether k8s metrics server is installed or not.
      if (!k8s_is_metrics_enabled($cloud_context)) {
        $this->messenger->addWarning($this->t('@message Please install %metrics_server_link into @cloud_context', [
          '@message' => $this->t('Node Resource Store requires K8s Metrics Server.'),
          '%metrics_server_link' => $this->k8sService->getMetricsServerLink($cloud_context),
          '@cloud_context' => $cloud_context,
        ]));
        continue;
      }

      $namespace_entities = $this->entityTypeManager
        ->getStorage('k8s_namespace')
        ->loadByProperties([
          'cloud_context' => $cloud_context,
        ]);
      if (empty($namespace_entities)) {
        continue;
      }

      foreach ($namespace_entities ?: [] as $namespace) {
        // Get resource metrics from each namespace.
        $queue->createItem([
          'k8s_method_name' => 'updateK8sNamespaceResourceStoreEntity',
          'cloud_context' => $cloud_context,
          'namespace' => $namespace->getName(),
          // @todo Ideally, we should use cost_type on worker node(s).
          // However, there is no way to get it right now.
          // Thus, temporary value is put here.
          'cost_type' => 'on_demand_yearly',
        ]);
      }
    }
  }

  /**
   * Extract data based on time range.
   *
   * @param string $cloud_store_type
   *   Cloud store type.
   * @param string $cloud_context
   *   The cloud context.
   * @param string $name
   *   The name of the cost store.
   * @param int $from_time
   *   The unix timestamp of the starting time.
   * @param int $to_time
   *   The unix timestamp of the ending time.
   * @param string $cost_type
   *   The cost type.
   *
   * @return array
   *   The result including resources, cost and unix time.
   */
  public function extract($cloud_store_type, $cloud_context, $name, $from_time, $to_time, $cost_type = NULL): array {
    $storage = $this->entityTypeManager->getStorage('cloud_store');
    $query = $storage->getQuery();
    $query->condition('type', $cloud_store_type);
    if (!empty($cloud_context)) {
      $query->condition('field_cloud_context', $cloud_context);
    }
    if (!empty($name)) {
      $query->condition('name', $name);
    }
    if (!empty($cost_type)) {
      $query->condition('field_cost_type', $cost_type);
    }
    $query->condition('type', $cloud_store_type);
    $query->sort('created');
    $entity_ids = $query->execute();
    $entities = $storage->loadMultiple($entity_ids);

    $result = [];
    $total_cost = 0;
    foreach ($entities ?: [] as $entity) {
      $created_time = $entity->get('created')->value;
      if ($to_time < $created_time || $created_time <= $from_time) {
        continue;
      }
      $resources = Yaml::decode(!empty($entity->get('field_resources')->value) ? $entity->get('field_resources')->value : '');
      $result[] = [
        'resources' => $resources,
        'name' => $entity->getName(),
        'created' => $created_time,
      ];
      if ($entity->hasField('field_cost')) {
        $cost = $entity->get('field_cost')->value;
        $result[count($result) - 1]['cost'] = $cost;
        $total_cost += $cost;
      }
    }
    $result['total_cost'] = $total_cost;
    return $result;
  }

  /**
   * {@inheritdoc}
   */
  public function updateCostStoreEntity(): void {
    $entity_type_manager = $this->entityTypeManager;
    $k8s_service = $this->k8sService;
    // Get target entities to be updated.
    $project_entities = $this->entityTypeManager
      ->getStorage('cloud_project')
      ->loadByProperties([]);

    if (empty($project_entities)) {
      return;
    }

    foreach ($project_entities ?: [] as $project_entity) {
      $label = $project_entity->label();
      $k8s_clusters = $project_entity->get('field_k8s_clusters');

      // Check whether namespace exists or not.
      $namespace_entities = $this->entityTypeManager->getStorage('k8s_namespace')
        ->loadByProperties([
          'name' => $label,
        ]);
      if (empty($namespace_entities)) {
        continue;
      }

      // Get latest entity.
      $store_entities = $this->entityTypeManager->getStorage('cloud_store')
        ->loadByProperties([
          'name' => $label,
          'type' => 'k8s_cost_store',
        ]);
      $store_entity = array_pop($store_entities);

      // To check time difference between current and latest update time
      // in order to decide whether to update or not.
      $latest_time = time() - (60 * 60 * 24);
      if (!empty($store_entity)) {
        $latest_time = $store_entity->get('created')->value;
        $time_diff = time() - $latest_time;

        // If previous update is within the specific range,
        // update process will be skipped.
        if ($time_diff < $this->length) {
          continue;
        }
      }

      $result[$label] = [];
      foreach ($k8s_clusters ?: [] as $k8s_cluster) {
        $cloud_context = $k8s_cluster->value;
        $current_time = floor(time() / $this->length) * $this->length;
        $ids = $this->entityTypeManager
          ->getStorage('cloud_store')
          ->getQuery()
          ->condition('type', 'k8s_namespace_resource_store')
          ->condition('field_cloud_context', $cloud_context)
          ->condition('name', $label)
          ->condition('created', [$latest_time, $current_time], 'BETWEEN')
          ->execute();
        $resource_entities = $this->entityTypeManager->getStorage('cloud_store')
          ->loadMultiple($ids);

        // If no resource entity, it tries to catch the resource info from k8s.
        if (empty($resource_entities)) {
          $namespace_entities = $this->entityTypeManager
            ->getStorage('k8s_namespace')->loadByProperties([
              'cloud_context' => $cloud_context,
              'name' => $label,
            ]);
          foreach ($namespace_entities ?: [] as $namespace_entity) {
            $namespace = $namespace_entity->getName();
            $k8s_service->updateK8sNamespaceResourceStoreEntity($cloud_context, $namespace, 'on_demand_yearly');
          }
          continue;
        }

        // Count cost in specific time range from all k8s clusters.
        foreach ($resource_entities ?: [] as $resource_entity) {
          $created_time = $resource_entity->get('created')->value;
          $time_range = ceil($created_time / $this->length) * $this->length;
          if (!isset($result[$label][$time_range])) {
            $result[$label][$time_range] = [
              'total_cost' => 0,
              'cost_type' => '',
              'count' => 0,
              'resources' => [
                'cpu_usage' => 0,
                'memory_usage' => 0,
                'pod_usage' => 0,
              ],
            ];
          }
          $result[$label][$time_range]['total_cost'] += $resource_entity->get('field_cost')->value;
          $result[$label][$time_range]['cost_type'] = $resource_entity->get('field_cost_type')->value;
          $result[$label][$time_range]['count'] += 1;
          $resources = Yaml::decode($resource_entity->get('field_resources')->value);
          $result[$label][$time_range]['resources']['cpu_usage'] += $resources['cpu_usage'];
          $result[$label][$time_range]['resources']['memory_usage'] += $resources['memory_usage'];
          $result[$label][$time_range]['resources']['pod_usage'] += $resources['pod_usage'];
        }
      }

      // Update cost entity.
      foreach ($result[$label] ?: [] as $key => $value) {
        $avg_cost = 0;
        $cpu_usage_avg = 0;
        $memory_usage_avg = 0;
        $pod_usage_avg = 0;
        if ($value['total_cost'] !== 0) {
          $avg_cost = (float) ($value['total_cost'] / $value['count']);
        }
        if ($value['resources']['cpu_usage'] > 0) {
          $cpu_usage_avg = (float) ($value['resources']['cpu_usage'] / $value['count']);
        }
        if ($value['resources']['memory_usage'] > 0) {
          $memory_usage_avg = (float) ($value['resources']['memory_usage'] / $value['count']);
        }
        if ($value['resources']['pod_usage'] > 0) {
          $pod_usage_avg = (float) ($value['resources']['pod_usage'] / $value['count']);
        }
        $resources = [
          'cpu_usage_avg' => $cpu_usage_avg,
          'memory_usage_avg' => $memory_usage_avg,
          'pod_usage_avg' => $pod_usage_avg,
          'count' => $value['count'],
        ];

        $data = [
          'name' => $label,
          'type' => 'k8s_cost_store',
          'field_cost_type' => $value['cost_type'],
          'field_cost' => $avg_cost,
          'field_resources' => Yaml::encode($resources),
          'created' => $key,
        ];
        // Update cost store entity.
        $entity = $entity_type_manager->getStorage('cloud_store')->create($data);
        $entity->save();
      }
    }
  }

}
